#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "repositoryview.h"
#include <QFileDialog>
#include <QMessageBox>
#include <QSettings>

static const char* SETTINGS_RECENT_FILE_LIST = "recentFileList";

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    refreshViewState();
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::on_actionCreate_repository_triggered()
{
    // TODO
}

void MainWindow::on_actionOpen_repository_triggered()
{
    auto dir = QFileDialog::getExistingDirectory(this, tr("Select repository"));
    if (dir.length() == 0) {
        return;
    }

    viewRepository(dir);
}

void MainWindow::viewRepository(const QString& path)
{
    // Open repo
    git_repository* gr = nullptr;
    int err = git_repository_open(&gr, path.toLocal8Bit().data());
    if (err != 0) {
        handleLibGitError(err);
        return;
    }

    QString shortName = QDir(path).dirName();

    auto *rv = new RepositoryView(gr, shortName, nullptr);

    // embed it into the view
    ui->tabWidget->addTab(rv, shortName);

    // switch to newest tab
    ui->tabWidget->setCurrentIndex( ui->tabWidget->count() - 1 );

    addToRecentRepositories(path); // calls refreshViewState();
}

void MainWindow::addToRecentRepositories(const QString& path)
{
    QSettings settings;
    auto rfl = settings.value(SETTINGS_RECENT_FILE_LIST).toStringList();

    // remove any existing entries that match it
    for(auto i = rfl.begin(); i != rfl.end();) {
        if (path == *i) {
            i = rfl.erase(i);
        } else {
            ++i;
        }
    }

    // insert at head
    rfl.push_front(path);

    // truncate limit
    while (rfl.length() > 10) {
        rfl.removeLast();
    }

    // persist
    settings.setValue(SETTINGS_RECENT_FILE_LIST, rfl);

    refreshViewState();
}

void MainWindow::refreshViewState()
{
    auto *cw = ui->tabWidget->currentWidget();

    bool anyRepoOpen = (cw != nullptr);
    ui->actionClose_repository->setEnabled( anyRepoOpen );
    ui->actionCommit->setEnabled( anyRepoOpen );
    ui->actionRefresh->setEnabled( anyRepoOpen );
    ui->actionRevision_Details->setEnabled( anyRepoOpen );
    ui->actionSynchronise->setEnabled( anyRepoOpen );

    if (cw == nullptr) {
        setWindowTitle("");
    } else {
        RepositoryView *rv = static_cast<RepositoryView*>(cw);
        setWindowTitle( rv->getShortName() );
    }

    // handle Recent Repositories list
    ui->menuRecent->clear();
    ui->menuRecent->addAction(ui->actionClear_recent_history);
    ui->menuRecent->addSeparator();
    QSettings settings;
    auto rfl = settings.value(SETTINGS_RECENT_FILE_LIST).toStringList();
    for (auto i = rfl.begin(); i != rfl.end(); ++i) {
        QAction *ac = new QAction();
        QString path = *i; // copy
        ac->setText(path);
        connect(ac, &QAction::triggered, [=]() {
            viewRepository(path);
        });
        ui->menuRecent->addAction(ac);
    }
    //ui->menuRecent->setVisible(true);
}

void MainWindow::handleLibGitError(int err)
{
    const git_error *e = giterr_last();
    auto *qmb = new QMessageBox();
    qmb->setText(tr("Error %1/%2: %3").arg(err).arg(e->klass).arg(e->message));
    connect(qmb, &QMessageBox::finished, qmb, &QMessageBox::deleteLater);
    qmb->show();
}

void MainWindow::on_actionExit_triggered()
{
    close();
}

void MainWindow::on_actionRefresh_triggered()
{
    auto *cw = ui->tabWidget->currentWidget();
    if (cw == nullptr) {
        return;
    }
    RepositoryView *rv = static_cast<RepositoryView*>(cw);
    rv->refreshView();
}

void MainWindow::on_actionRevision_Details_triggered()
{
    auto *cw = ui->tabWidget->currentWidget();
    if (cw == nullptr) {
        return;
    }
    RepositoryView *rv = static_cast<RepositoryView*>(cw);
    rv->viewInfoPage();
}

void MainWindow::on_actionCommit_triggered()
{
    auto *cw = ui->tabWidget->currentWidget();
    if (cw == nullptr) {
        return;
    }
    RepositoryView *rv = static_cast<RepositoryView*>(cw);
    rv->viewCommitPage();
}

void MainWindow::on_actionSynchronise_triggered()
{
    auto *cw = ui->tabWidget->currentWidget();
    if (cw == nullptr) {
        return;
    }
    RepositoryView *rv = static_cast<RepositoryView*>(cw);
    rv->viewSyncPage();
}

void MainWindow::on_tabWidget_currentChanged(int index)
{
    Q_UNUSED(index);
    refreshViewState();
}

void MainWindow::on_tabWidget_tabCloseRequested(int index)
{
    // QTabWidget won't remove the tab by itself, we have to do it
    // This deletes the QWidget* RepositoryView*, which will clean up the underlying repository connection too
    ui->tabWidget->removeTab(index);
}

void MainWindow::on_actionClose_repository_triggered()
{
    int idx = ui->tabWidget->currentIndex();
    if (idx >= 0) {
        ui->tabWidget->removeTab(idx);
    }
}

void MainWindow::on_actionClear_recent_history_triggered()
{
    QSettings settings;
    settings.remove(SETTINGS_RECENT_FILE_LIST);
    refreshViewState();
}
