#include "repositoryview.h"
#include "ui_repositoryview.h"
#include "mainwindow.h"
#include "graphhistory.h"
#include <QDateTime>

static const int RevisionShaRole = Qt::UserRole + 1;
static const int DiffInfoRole    = Qt::UserRole + 2;

RepositoryView::RepositoryView(git_repository *r, const QString& shortName, QWidget *parent) :
    QWidget(parent),
    ui(new Ui::RepositoryView),
    r(r),
    lastStatus(nullptr),
    shortName(shortName)
{
    ui->setupUi(this);

    // always start on Info page
    ui->stackedWidget->setCurrentIndex(0);

    ui->commitHistoryTree->header()->setStretchLastSection(false);
    ui->commitHistoryTree->header()->setSectionResizeMode(2, QHeaderView::Stretch);

    ui->stageFilesTree->header()->setStretchLastSection(false);
    ui->stageFilesTree->header()->setSectionResizeMode(0, QHeaderView::Stretch);
    ui->stageFilesTree->header()->setSectionResizeMode(1, QHeaderView::ResizeToContents);

    refreshView();
}

RepositoryView::~RepositoryView()
{
    git_repository_free(r);
    delete ui;
}

const QString& RepositoryView::getShortName()
{
    return this->shortName;
}

void RepositoryView::refreshView()
{
    ui->commitHistoryTree->clear();

    git_revwalk* walker = nullptr;
    int err = git_revwalk_new(&walker, this->r);
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    // sort children after parents, newest commits first
    git_revwalk_sorting(walker, GIT_SORT_TOPOLOGICAL | GIT_SORT_TIME);

    // push all refs, not just HEAD
    err = git_revwalk_push_glob(walker, "*"); // git_revwalk_push_head(walker);
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        git_revwalk_free(walker);
        return;
    }

    QVector<git_oid> lines;
    GraphHistory graph;

    git_oid oid;
    while (!git_revwalk_next(&oid, walker)) {

        git_commit *commit;
        err = git_commit_lookup(&commit, this->r, &oid);
        if (err != 0) {
            MainWindow::handleLibGitError(err);
            git_revwalk_free(walker);
        }

        // Only display first line of the commit message
        QString commitMessage(git_commit_message(commit));
        int lf = commitMessage.indexOf('\n');
        if (lf >= 0) {
            commitMessage.resize(lf);
        }

        QTreeWidgetItem *itm = new QTreeWidgetItem();
        itm->setText(2, commitMessage);
        itm->setText(3, QString(git_commit_author(commit)->name));

        char fullsha[GIT_OID_HEXSZ+1] = {0};
        git_oid_tostr(fullsha, GIT_OID_HEXSZ+1, &oid);
        itm->setText(4, QString(fullsha).mid(0, 8));

        QString qFullSha = QString(fullsha);
        itm->setData(0, RevisionShaRole, qFullSha);

        itm->setText(0, graph.addRow(commit));

        ui->commitHistoryTree->addTopLevelItem(itm);

        git_commit_free(commit);
    }

    git_revwalk_free(walker);

    for (int i = ui->commitHistoryTree->columnCount(); i --> 0;) {
        ui->commitHistoryTree->resizeColumnToContents(i);
    }
}

void RepositoryView::viewInfoPage()
{
    ui->stackedWidget->setCurrentIndex(0);
}

void RepositoryView::viewCommitPage()
{
    ui->stackedWidget->setCurrentIndex(1);
    refreshStatus();
}

void RepositoryView::viewSyncPage()
{
    ui->stackedWidget->setCurrentIndex(2);
}

void RepositoryView::refreshStatus()
{
    // Clean up

    ui->stageFilesTree->clear();

    if (lastStatus != nullptr) {
        git_status_list_free(lastStatus);
    }

    //

    git_status_options opts = GIT_STATUS_OPTIONS_INIT;
    opts.flags |= GIT_STATUS_OPT_INCLUDE_UNTRACKED;
    opts.flags |= GIT_STATUS_OPT_RECURSE_UNTRACKED_DIRS;
    opts.flags |= GIT_STATUS_OPT_EXCLUDE_SUBMODULES; // ?? should we use this by default?
    opts.show = GIT_STATUS_SHOW_INDEX_AND_WORKDIR;
    //opts.flags |= GIT_STATUS_OPT_RENAMES_HEAD_TO_INDEX; // rename detection

    // @ref https://github.com/jacquesg/p5-Git-Raw/issues/29#issuecomment-31974158
    // @ref https://libgit2.github.com/libgit2/ex/HEAD/status.html#section-7

    git_status_list *status = nullptr;
    int err = git_status_list_new(&status, this->r, &opts);
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    int nEntries = git_status_list_entrycount(status);
    for (int i = 0; i < nEntries; ++i) {

       const git_status_entry *s = git_status_byindex(status, i);

       if (s->status == GIT_STATUS_CURRENT || s->status == GIT_STATUS_INDEX_TYPECHANGE) {
           continue;
       }

       auto *itm = new QTreeWidgetItem();

       const char* new_path = nullptr;
       git_off_t filesize = 0;

       if (s->head_to_index != nullptr) {
           itm->setData(0, DiffInfoRole, QVariant::fromValue(intptr_t(s->head_to_index)));
           filesize = s->head_to_index->new_file.size;
           new_path = s->head_to_index->new_file.path;

       } else if (s->index_to_workdir != nullptr) {
           itm->setData(0, DiffInfoRole, QVariant::fromValue(intptr_t(s->index_to_workdir)));
           filesize = s->index_to_workdir->new_file.size;
           new_path = s->index_to_workdir->new_file.path;

       } else {
           itm->setData(0, DiffInfoRole, QVariant::fromValue(0));
           filesize = 0;
           new_path = "???";

       }

       itm->setText(0, QString(new_path));
       itm->setText(1, QString("%1").arg(filesize));
       itm->setCheckState(0, Qt::Unchecked);
       ui->stageFilesTree->addTopLevelItem(itm);

    }

    lastStatus = status; // don't free it - used for viewing diffs, etc
}

void RepositoryView::on_commitHistoryTree_currentItemChanged(QTreeWidgetItem *current, QTreeWidgetItem *previous)
{
    Q_UNUSED(previous);

    if (current == nullptr) {
        // TODO clear selection

    } else {
        viewInfoFor(current->data(0, RevisionShaRole).toString());

    }
}

void RepositoryView::viewInfoFor(const QString& revisionSha)
{
    viewInfoPage();
    git_oid oid = {0};
    int err = git_oid_fromstr(&oid, revisionSha.toLocal8Bit().data());
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    git_commit *commit;
    err = git_commit_lookup(&commit, this->r, &oid);
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    ui->fileInfo->clear();
    ui->commitMessage->clear();
    ui->fileDiff->clear();

    ui->commitMessage->setPlainText( git_commit_message(commit) );
    ui->commitRevision->setText( revisionSha );

    auto sig = git_commit_author(commit);
    ui->commitAuthor->setText(tr("%1 <%2>").arg(sig->name).arg(sig->email));

    QDateTime qdt = QDateTime::fromTime_t(sig->when.time);
    qdt.setOffsetFromUtc(sig->when.offset * 60);
    ui->commitDate->setText(qdt.toString());

}

void RepositoryView::on_commitRefreshView_clicked()
{
    refreshStatus();
}

void RepositoryView::on_stageFilesTree_currentItemChanged(QTreeWidgetItem *current, QTreeWidgetItem *previous)
{
    Q_UNUSED(previous);

    ui->fileDiffStagedDisplay->clear();
    if (current == nullptr) {
        ui->fileDiffStagedDisplay->setEnabled(false); // selection cleared
        return;
    }

    auto* gdd = (git_diff_delta*)( intptr_t(current->data(0, DiffInfoRole).toLongLong()) );
    if (gdd == nullptr) {
        ui->fileDiffStagedDisplay->setEnabled(false); // some kind of error
        return;
    }

    ui->fileDiffStagedDisplay->setEnabled(true);
    ui->fileDiffStagedDisplay->setPlainText(QString("got diff info for %1").arg(gdd->new_file.path));

}

void RepositoryView::on_pushButton_clicked()
{
    commit();
}

void RepositoryView::commit()
{
    int nParents = 0;
    git_commit *parent = nullptr;

    int err = git_repository_head_unborn(this->r);
    if (err == 1) {
        // head is unborn - need to create it before committing the index

    } else if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;

    } else {
        // resolve HEAD to a commit hash
        git_oid oidHead = {0};
        int err2 = git_reference_name_to_id( &oidHead, this->r, "HEAD" );
        if (err2 != 0) {
            MainWindow::handleLibGitError(err);
            return;
        }

        // resolve commit hash to git_commit* object
        err2 = git_commit_lookup(&parent, this->r, &oidHead);
        if (err2 != 0) {
            MainWindow::handleLibGitError(err);
            return;
        }

        nParents = 1;
    }

    // @ref https://stackoverflow.com/a/27673192

    // get the staging index
    git_index *index = nullptr;
    err = git_repository_index(&index, this->r);
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    // make a note of every path that's selected
    QStringList selectedPaths;
    for (int i = 0; i < ui->stageFilesTree->topLevelItemCount(); ++i) {
        auto *itm = ui->stageFilesTree->topLevelItem(i);

        if (itm->checkState(0) == Qt::Checked) {
            selectedPaths.push_back( ui->stageFilesTree->topLevelItem(i)->text(0) );
        }
    }

    // reset the staging index to match head
    // @ref https://github.com/libgit2/libgit2/issues/3632#issuecomment-187934447
    /*
    git_reference *head;
    git_object *head_commit;
    git_repository_head(&head, this->r);
    git_reference_peel(&head_commit, head, GIT_OBJ_COMMIT);
    git_reset_default(this->r, head_commit, &paths);
    */

    // re-add to the staging index only by selected paths
    for (int i = 0; i < selectedPaths.size(); ++i) {
        auto buff = selectedPaths[i].toUtf8();
        err = git_index_add_bypath(index, buff.data());
        if (err != 0) {
            MainWindow::handleLibGitError(err);
            return;
        }
    }

    // write the staging index to a tree
    git_oid tree_id = {0};
    err = git_index_write_tree(&tree_id, index);
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    git_tree* tree = nullptr;
    err = git_tree_lookup(&tree, this->r, &tree_id);
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    // Load author + committer information
    // FIXME load from user.XXX settings
    git_signature *author, *committer;
    git_signature_new(&author, "Author", "author@example.com", 0, 0);
    git_signature_new(&committer, "Committer", "committer@example.com", 0, 0);

    // Commit message
    auto commitMessage = ui->editCommitMessage->toPlainText().toUtf8();

    // Commit
    git_oid commit_id = {0};
    err = git_commit_create_v(
        &commit_id,
        this->r,
        "HEAD",     /* The commit will update the position of HEAD */
        author,
        committer,
        NULL,       /* UTF-8 encoding */
        commitMessage.data(),
        tree,       /* The tree from the index */

        // parents
        // (could have used create - non_v instead of passing 0 + extra argument)
        nParents,
        parent
    );
    if (err != 0) {
        MainWindow::handleLibGitError(err);
        return;
    }

    // clean up
    git_index_free(index);
    git_tree_free(tree);
    git_signature_free(author);
    git_signature_free(committer);
    if (parent != nullptr) {
        git_commit_free(parent);
    }
    ui->editCommitMessage->clear();

    refreshStatus();
}
